@@ -1,32 +1,44 @@
-# -*- coding: utf-8 -*-
-# Copyright (C) 2009-2013  Roman Zimbelmann <hut@lavabit.com>
-# This software is distributed under the terms of the GNU GPL version 3.
+# This file is part of ranger, the console file manager.
+# License: GNU GPL version 3, see the file "AUTHORS" for details.
 
 """The BrowserColumn widget displays the contents of a directory or file."""
 
+from __future__ import (absolute_import, division, print_function)
+
 import curses
 import stat
 from time import time
+from os.path import splitext
+
+try:
+    from bidi.algorithm import get_display  # pylint: disable=import-error
+    HAVE_BIDI = True
+except ImportError:
+    HAVE_BIDI = False
+
+from ranger.ext.widestring import WideString
+from ranger.core import linemode
 
 from . import Widget
 from .pager import Pager
-from ranger.ext.widestring import WideString
 
-from ranger.gui.color import *
 
-class BrowserColumn(Pager):
+def hook_before_drawing(fsobject, color_list):
+    return fsobject, color_list
+
+
+class BrowserColumn(Pager):  # pylint: disable=too-many-instance-attributes
     main_column = False
     display_infostring = False
-    display_vcsstate   = True
+    display_vcsstate = True
     scroll_begin = 0
     target = None
     last_redraw_time = -1
-    ellipsis = { False: '~', True: '…' }
 
     old_dir = None
     old_thisfile = None
 
-    def __init__(self, win, level):
+    def __init__(self, win, level, tab=None):
         """Initializes a Browser Column Widget
 
         win = the curses window object of the BrowserView
@@ -36,21 +48,22 @@ class BrowserColumn(Pager):
         level 0 => current file/directory
         level <0 => parent directories
         """
+        self.need_redraw = False
+        self.image = None
+        self.need_clear_image = True
         Pager.__init__(self, win)
-        Widget.__init__(self, win)
+        Widget.__init__(self, win)  # pylint: disable=non-parent-init-called
         self.level = level
+        self.tab = tab
         self.original_level = level
 
         self.settings.signal_bind('setopt.display_size_in_main_column',
-                self.request_redraw, weak=True)
+                                  self.request_redraw, weak=True)
 
     def request_redraw(self):
         self.need_redraw = True
 
-    def resize(self, y, x, hei, wid):
-        Widget.resize(self, y, x, hei, wid)
-
-    def click(self, event):
+    def click(self, event):     # pylint: disable=too-many-branches
         """Handle a MouseEvent"""
         direction = event.mouse_wheel_direction()
         if not (event.pressed(1) or event.pressed(3) or direction):
@@ -77,14 +90,16 @@ class BrowserColumn(Pager):
                 elif event.pressed(3):
                     try:
                         clicked_file = self.target.files[index]
+                    except IndexError:
+                        pass
+                    else:
                         if clicked_file.is_directory:
-                            self.fm.enter_dir(clicked_file.path)
+                            self.fm.enter_dir(clicked_file.path, remember=True)
                         elif self.level == 0:
                             self.fm.thisdir.move_to_obj(clicked_file)
                             self.fm.execute_file(clicked_file)
-                    except:
-                        pass
-
+        elif self.target.is_file:
+            self.scrollbit(direction)
         else:
             if self.level > 0 and not direction:
                 self.fm.move(right=0)
@@ -103,7 +118,7 @@ class BrowserColumn(Pager):
         """
         try:
             self.win.move(line, 0)
-        except:
+        except curses.error:
             return
         for entry in commands:
             text, attr = entry
@@ -131,36 +146,46 @@ class BrowserColumn(Pager):
 
     def poke(self):
         Widget.poke(self)
-        self.target = self.fm.thistab.at_level(self.level)
+        if self.tab is None:
+            tab = self.fm.thistab
+        else:
+            tab = self.tab
+        self.target = tab.at_level(self.level)
 
     def draw(self):
         """Call either _draw_file() or _draw_directory()"""
-        if self.target != self.old_dir:
+        target = self.target
+
+        if target != self.old_dir:
             self.need_redraw = True
-            self.old_dir = self.target
+            self.old_dir = target
+            self.scroll_extra = 0  # reset scroll start
 
-        if self.target:     # don't garbage collect this directory please
-            self.target.use()
+        if target:
+            target.use()
 
-        if self.target and self.target.is_directory \
-                and (self.level <= 0 or self.settings.preview_directories):
-            if self.target.pointed_obj != self.old_thisfile:
-                self.need_redraw = True
-                self.old_thisfile = self.target.pointed_obj
-
-            if self.target.load_content_if_outdated() \
-            or self.target.sort_if_outdated() \
-            or self.last_redraw_time < self.target.last_update_time:
-                self.need_redraw = True
+            if target.is_directory and (self.level <= 0 or self.settings.preview_directories):
+                if self.old_thisfile != target.pointed_obj:
+                    self.old_thisfile = target.pointed_obj
+                    self.need_redraw = True
+                self.need_redraw |= target.load_content_if_outdated()
+                self.need_redraw |= target.sort_if_outdated()
+                self.need_redraw |= self.last_redraw_time < target.last_update_time
+                if target.pointed_obj:
+                    self.need_redraw |= target.pointed_obj.load_if_outdated()
+                    self.need_redraw |= self.last_redraw_time < target.pointed_obj.last_load_time
+            else:
+                self.need_redraw |= target.load_if_outdated()
+                self.need_redraw |= self.last_redraw_time < target.last_load_time
 
         if self.need_redraw:
             self.win.erase()
-            if self.target is None:
+            if target is None:
                 pass
-            elif self.target.is_file:
+            elif target.is_file:
                 Pager.open(self)
                 self._draw_file()
-            elif self.target.is_directory:
+            elif target.is_directory:
                 self._draw_directory()
                 Widget.draw(self)
             self.need_redraw = False
@@ -169,27 +194,41 @@ class BrowserColumn(Pager):
     def _draw_file(self):
         """Draw a preview of the file, if the settings allow it"""
         self.win.move(0, 0)
+        if self.target is None or not self.target.has_preview():
+            Pager.close(self)
+            return
+
         if not self.target.accessible:
             self.addnstr("not accessible", self.wid)
             Pager.close(self)
             return
 
-        if self.target is None or not self.target.has_preview():
+        path = self.target.get_preview_source(self.wid, self.hei)
+        if path is None:
             Pager.close(self)
-            return
-
-        if self.fm.settings.preview_images and self.target.image:
-            self.set_image(self.target.realpath)
-            Pager.draw(self)
         else:
-            f = self.target.get_preview_source(self.wid, self.hei)
-            if f is None:
-                Pager.close(self)
+            if self.target.is_image_preview():
+                self.set_image(path)
             else:
-                self.set_source(f)
-                Pager.draw(self)
+                self.set_source(path)
+            Pager.draw(self)
 
-    def _draw_directory(self):
+    def _format_line_number(self, linum_format, i, selected_i):
+        line_number = i
+        if self.settings.line_numbers == 'relative':
+            line_number = abs(selected_i - i)
+            if not self.settings.relative_current_zero and line_number == 0:
+                if self.settings.one_indexed:
+                    line_number = selected_i + 1
+                else:
+                    line_number = selected_i
+        elif self.settings.one_indexed:
+            line_number += 1
+
+        return linum_format.format(line_number)
+
+    def _draw_directory(  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
+            self):
         """Draw the contents of a directory"""
         if self.image:
             self.image = None
@@ -201,6 +240,15 @@ class BrowserColumn(Pager):
 
         base_color = ['in_browser']
 
+        if self.fm.ui.viewmode == 'multipane' and self.tab is not None:
+            active_pane = self.tab == self.fm.thistab
+            if active_pane:
+                base_color.append('active_pane')
+            else:
+                base_color.append('inactive_pane')
+        else:
+            active_pane = False
+
         self.win.move(0, 0)
 
         if not self.target.content_loaded:
@@ -228,11 +276,16 @@ class BrowserColumn(Pager):
 
         copied = [f.path for f in self.fm.copy_buffer]
 
-        selected_i = self.target.pointer
+        # Set the size of the linum text field to the number of digits in the
+        # visible files in directory.
+        linum_text_len = len(str(self.scroll_begin + self.hei))
+        linum_format = "{0:>" + str(linum_text_len) + "}"
+        # add separator between line number and tag
+        linum_format += " "
+
+        selected_i = self._get_index_of_selected_file()
         for line in range(self.hei):
             i = line + self.scroll_begin
-            if line > self.hei:
-                break
 
             try:
                 drawn = self.target.files[i]
@@ -245,18 +298,39 @@ class BrowserColumn(Pager):
             else:
                 tagged_marker = " "
 
-            key = (self.wid, selected_i == i, drawn.marked, self.main_column,
-                    drawn.path in copied, tagged_marker, drawn.infostring,
-                    drawn.vcsfilestatus, drawn.vcsremotestatus, self.fm.do_cut)
+            # Extract linemode-related information from the drawn object
+            metadata = None
+            current_linemode = drawn.linemode_dict[drawn.linemode]
+            if current_linemode.uses_metadata:
+                metadata = self.fm.metadata.get_metadata(drawn.path)
+                if not all(getattr(metadata, tag)
+                           for tag in current_linemode.required_metadata):
+                    current_linemode = drawn.linemode_dict[linemode.DEFAULT_LINEMODE]
 
+            metakey = hash(repr(sorted(metadata.items()))) if metadata else 0
+            key = (self.wid, selected_i == i, drawn.marked, self.main_column,
+                   drawn.path in copied, tagged_marker, drawn.infostring,
+                   drawn.vcsstatus, drawn.vcsremotestatus, self.target.has_vcschild,
+                   self.fm.do_cut, current_linemode.name, metakey, active_pane,
+                   self.settings.line_numbers)
+
+            # Check if current line has not already computed and cached
             if key in drawn.display_data:
+                # Recompute line numbers because they can't be reliably cached.
+                if self.main_column and self.settings.line_numbers != 'false':
+                    line_number_text = self._format_line_number(linum_format,
+                                                                i,
+                                                                selected_i)
+                    drawn.display_data[key][0][0] = line_number_text
+
                 self.execute_curses_batch(line, drawn.display_data[key])
                 self.color_reset()
                 continue
 
-            text = drawn.basename
-            if drawn.marked and (self.main_column or \
-                    self.settings.display_tags_in_all_columns):
+            text = current_linemode.filetitle(drawn, metadata)
+
+            if drawn.marked and (self.main_column
+                                 or self.settings.display_tags_in_all_columns):
                 text = " " + text
 
             # Computing predisplay data. predisplay contains a list of lists
@@ -267,6 +341,19 @@ class BrowserColumn(Pager):
             predisplay_right = []
             space = self.wid
 
+            # line number field
+            if self.settings.line_numbers != 'false':
+                if self.main_column and space - linum_text_len > 2:
+                    line_number_text = self._format_line_number(linum_format,
+                                                                i,
+                                                                selected_i)
+                    predisplay_left.append([line_number_text, []])
+                    space -= linum_text_len
+
+                    # Delete one additional character for space separator
+                    # between the line number and the tag
+                    space -= 1
+
             # selection mark
             tagmark = self._draw_tagged_display(tagged, tagged_marker)
             tagmarklen = self._total_len(tagmark)
@@ -282,31 +369,41 @@ class BrowserColumn(Pager):
                 space -= vcsstringlen
 
             # info string
-            infostring = self._draw_infostring_display(drawn, space)
-            infostringlen = self._total_len(infostring)
-            if space - infostringlen > 2:
-                predisplay_right = infostring + predisplay_right
-                space -= infostringlen
+            infostring = []
+            infostringlen = 0
+            try:
+                infostringdata = current_linemode.infostring(drawn, metadata)
+                if infostringdata:
+                    infostring.append([" " + infostringdata + " ",
+                                       ["infostring"]])
+            except NotImplementedError:
+                infostring = self._draw_infostring_display(drawn, space)
+            if infostring:
+                infostringlen = self._total_len(infostring)
+                if space - infostringlen > 2:
+                    predisplay_right = infostring + predisplay_right
+                    space -= infostringlen
 
             textstring = self._draw_text_display(text, space)
             textstringlen = self._total_len(textstring)
             predisplay_left += textstring
             space -= textstringlen
 
+            assert space >= 0, "Error: there is not enough space to write the text. " \
+                "I have computed spaces wrong."
             if space > 0:
                 predisplay_left.append([' ' * space, []])
-            elif space < 0:
-                raise Exception("Error: there is not enough space to write "
-                        "the text. I have computed spaces wrong.")
 
             # Computing display data. Now we compute the display_data list
             # ready to display in curses. It is a list of lists [string, attr]
 
             this_color = base_color + list(drawn.mimetype_tuple) + \
-                    self._draw_directory_color(i, drawn, copied)
+                self._draw_directory_color(i, drawn, copied)
             display_data = []
             drawn.display_data[key] = display_data
 
+            drawn, this_color = hook_before_drawing(drawn, this_color)
+
             predisplay = predisplay_left + predisplay_right
             for txt, color in predisplay:
                 attr = self.settings.colorscheme.get_attr(*(this_color + color))
@@ -315,12 +412,28 @@ class BrowserColumn(Pager):
             self.execute_curses_batch(line, display_data)
             self.color_reset()
 
-    def _total_len(self, predisplay):
-        return sum([len(WideString(s)) for s, L in predisplay])
+    def _get_index_of_selected_file(self):
+        if self.fm.ui.viewmode == 'multipane' and self.tab:
+            return self.tab.pointer
+        return self.target.pointer
+
+    @staticmethod
+    def _total_len(predisplay):
+        return sum([len(WideString(s)) for s, _ in predisplay])
+
+    def _bidi_transpose(self, text):
+        if self.settings.bidi_support and HAVE_BIDI:
+            return get_display(text)
+        return text
 
     def _draw_text_display(self, text, space):
-        wtext = WideString(text)
+        bidi_text = self._bidi_transpose(text)
+        wtext = WideString(bidi_text)
+        wext = WideString(splitext(bidi_text)[1])
         wellip = WideString(self.ellipsis[self.settings.unicode_ellipsis])
+        if len(wtext) > space:
+            wtext = wtext[:max(1, space - len(wext) - len(wellip))] + wellip + wext
+        # Truncate again if still too long.
         if len(wtext) > space:
             wtext = wtext[:max(0, space - len(wellip))] + wellip
 
@@ -347,30 +460,26 @@ class BrowserColumn(Pager):
 
     def _draw_vcsstring_display(self, drawn):
         vcsstring_display = []
-        if self.settings.vcs_aware and (drawn.vcsfilestatus or \
-                drawn.vcsremotestatus):
-            if drawn.vcsfilestatus:
-                vcsstr, vcscol = self.vcsfilestatus_symb[drawn.vcsfilestatus]
-            else:
-                vcsstr = " "
-                vcscol = []
-            vcsstring_display.append([vcsstr, ['vcsfile'] + vcscol])
-
+        if (self.target.vcs and self.target.vcs.track) \
+                or (drawn.is_directory and drawn.vcs and drawn.vcs.track):
             if drawn.vcsremotestatus:
-                vcsstr, vcscol = self.vcsremotestatus_symb[
-                        drawn.vcsremotestatus]
-            else:
-
-                vcsstr = " "
-                vcscol = []
-            vcsstring_display.append([vcsstr, ['vcsremote'] + vcscol])
+                vcsstr, vcscol = self.vcsremotestatus_symb[drawn.vcsremotestatus]
+                vcsstring_display.append([vcsstr, ['vcsremote'] + vcscol])
+            elif self.target.has_vcschild:
+                vcsstring_display.append([' ', []])
+            if drawn.vcsstatus:
+                vcsstr, vcscol = self.vcsstatus_symb[drawn.vcsstatus]
+                vcsstring_display.append([vcsstr, ['vcsfile'] + vcscol])
+            elif self.target.has_vcschild:
+                vcsstring_display.append([' ', []])
         elif self.target.has_vcschild:
-            vcsstring_display.append(["  ", []])
+            vcsstring_display.append(['  ', []])
+
         return vcsstring_display
 
     def _draw_directory_color(self, i, drawn, copied):
         this_color = []
-        if i == self.target.pointer:
+        if i == self._get_index_of_selected_file():
             this_color.append('selected')
 
         if drawn.marked:
@@ -404,13 +513,13 @@ class BrowserColumn(Pager):
 
         return this_color
 
-    def _get_scroll_begin(self):
+    def _get_scroll_begin(self):  # pylint: disable=too-many-return-statements
         """Determines scroll_begin (the position of the first displayed file)"""
         offset = self.settings.scroll_offset
         dirsize = len(self.target)
         winsize = self.hei
         halfwinsize = winsize // 2
-        index = self.target.pointer or 0
+        index = self._get_index_of_selected_file() or 0
         original = self.target.scroll_begin
         projected = index - original
 
@@ -424,7 +533,7 @@ class BrowserColumn(Pager):
             return 0
 
         if halfwinsize < offset:
-            return min( dirsize - winsize, max( 0, index - halfwinsize ))
+            return min(dirsize - winsize, max(0, index - halfwinsize))
 
         if original > dirsize - winsize:
             self.target.scroll_begin = dirsize - winsize
@@ -434,12 +543,10 @@ class BrowserColumn(Pager):
             return original
 
         if projected > upper_limit:
-            return min( dirsize - winsize,
-                    original + (projected - upper_limit))
+            return min(dirsize - winsize, original + (projected - upper_limit))
 
         if projected < upper_limit:
-            return max( 0,
-                    original - (lower_limit - projected))
+            return max(0, original - (lower_limit - projected))
 
         return original
 
